#include "vtkDotSource.h"

#include "vtkDoubleArray.h"
#include "vtkPolyData.h"
#include "vtkInformation.h"
#include "vtkInformationDoubleVectorKey.h"
#include "vtkInformationIntegerVectorKey.h"
#include "vtkInformationVector.h"
#include "vtkObjectFactory.h"
#include "vtkMultiProcessController.h"
#include "vtkPointData.h"
#include "vtkStreamingDemandDrivenPipeline.h"

#include <vtksys/ios/sstream>

vtkCxxRevisionMacro(vtkDotSource,"$Revision: 1283 $");
vtkInstantiatorNewMacro(vtkDotSource);
vtkCxxSetObjectMacro(vtkDotSource,Controller,vtkMultiProcessController);

vtkDotSource* vtkDotSource::Singleton = 0;

vtkDotSource* vtkDotSource::New()
{
  if ( vtkDotSource::Singleton )
    {
    return vtkDotSource::Singleton;
    }

  // This sets vtkDotSource::Singleton:
  vtkDotSource* src;
  vtkObject* ret = vtkObjectFactory::CreateInstance( "vtkDotSource" );
  if ( ret )
    {
    src = static_cast<vtkDotSource*>( ret );
    }
  else
    {
    src = new vtkDotSource;
    }
  // Add an extra reference so the singleton isn't destroyed
  // until vtkDotSource::DestroySingleton() is called:
  src->Register( 0 );
  return src;
}

vtkDotSource::vtkDotSource()
{
  this->Initialized = 0;
  this->Controller = 0;
  this->LocalId = 0;
  this->NumberOfProcesses = 1;
  this->LocalParticles[0] = this->LocalParticles[1] = -1;
  this->NumberOfLocalParticles = 0;
  this->NumberOfTotalParticles = 0;
  this->SimulationData = vtkPolyData::New();
  this->SetNumberOfInputPorts( 0 );
  this->SetNumberOfOutputPorts( 1 ); // Port 0: vtkPolyData

  vtkDotSource::Singleton = this;
}

vtkDotSource::~vtkDotSource()
{
  vtkDotSource::Singleton = 0;

  this->SimulationData->Delete();
  this->SetController( 0 );
}

void vtkDotSource::DestroySingleton()
{
  if ( vtkDotSource::Singleton )
    {
    int refCount = vtkDotSource::Singleton->GetReferenceCount();
    if ( refCount > 1 )
      {
      vtkGenericWarningMacro(
        << "Attempting to destroy the Dot singleton while there are "
        << (refCount - 1) << " outstanding references.\n"
        << "Destruction will be delayed until their release." );
      }
    vtkDotSource::Singleton->Delete();
    }
}

void vtkDotSource::PrintSelf( ostream& os, vtkIndent indent )
{
  this->Superclass::PrintSelf( os, indent );
  os << "Initialized: " << this->Initialized << "\n";
  os << "SimulationData: " << this->SimulationData << "\n";
  os << "SimulationTime: " << this->SimulationTime << "\n";
  os << "Controller: " << this->Controller << "\n";
  os << "LocalId: " << this->LocalId << "\n";
  os << "NumberOfProcesses: " << this->NumberOfProcesses << "\n";
  os << "LocalParticles: " << this->LocalParticles[0] << " -- " << this->LocalParticles[1] << "\n";
  os << "[Singleton]: " << vtkDotSource::Singleton << "\n";
}

void vtkDotSource::AddPointArray( double* ptr, int numComponents, const char* name )
{
  vtkPolyData* sim = this->SimulationData;
  vtkDoubleArray* arr = vtkDoubleArray::New();
  arr->SetNumberOfComponents( numComponents );
  //cout << "Adding " << name << " with " << sim->GetNumberOfPoints() << " tuples\n";
  arr->SetArray( ptr + numComponents * this->LocalParticles[0], numComponents * this->NumberOfLocalParticles, 1 );
  arr->SetName( name );
  sim->GetPointData()->AddArray( arr );
  arr->FastDelete();
  if ( numComponents == 1 )
    {
    sim->GetPointData()->SetScalars( arr );
    }
  else if ( numComponents == 3 )
    {
    sim->GetPointData()->SetVectors( arr );
    }
}

/// Prepare the singleton using data from the simulation.
void vtkDotSource::Initialize(
  int num_particles, double* pxyz, double* mass, double* pmom, double* pfrc, int local_id, int num_procs )
{
  if ( this->Initialized )
    {
    return;
    }
  vtkPolyData* sim = this->SimulationData;
  vtkInformation* simInfo = sim->GetInformation();
  // Clear any existing data:
  sim->Initialize();
  simInfo->Set( vtkDataObject::DATA_PIECE_NUMBER(), local_id );
  simInfo->Set( vtkDataObject::DATA_NUMBER_OF_PIECES(), num_procs );
  if ( num_procs > 1 )
    {
    this->SetController( vtkMultiProcessController::GetGlobalController() );
    }
  this->LocalId = local_id;
  this->NumberOfProcesses = num_procs;
  this->NumberOfTotalParticles = num_particles;
  this->UpdateLocalParticles();

  vtkDoubleArray* arr = vtkDoubleArray::New();
  arr->SetNumberOfComponents( 3 );
  arr->SetArray( pxyz + 3 * this->LocalParticles[0], 3 * this->NumberOfLocalParticles, 1 );
  arr->SetName( "Coordinates" );
  vtkPoints* pts = vtkPoints::New();
  pts->SetData( arr );
  sim->SetPoints( pts );
  arr->FastDelete();
  pts->FastDelete();
  this->AddPointArray( mass, 1, "Mass" );
  this->AddPointArray( pmom, 3, "Momentum" );
  this->AddPointArray( pfrc, 3, "Force" );
  this->Initialized = 1;
}

void vtkDotSource::UpdateSimulationTime( double time )
{
  // Unlike SetXXX() methods, this always marks the dataset as modified.
  // This is necessary since the simulation will have changed values in
  // the underlying arrays without VTK being involved, making a recompute
  // of the pipeline necessary.
  this->SimulationTime = time;
  this->Modified();
}

int vtkDotSource::RequestInformation( vtkInformation* request, vtkInformationVector** inInfoVec, vtkInformationVector* ouInfoVec )
{
  vtkInformation* ouInfo0 = ouInfoVec->GetInformationObject( 0 );
  ouInfo0->Set( vtkStreamingDemandDrivenPipeline::MAXIMUM_NUMBER_OF_PIECES(), -1 );
  ouInfo0->Set( vtkStreamingDemandDrivenPipeline::TIME_STEPS(), &this->SimulationTime, 1 );

  vtkPolyData* sim = this->SimulationData;
  if ( sim )
    {
    // FIXME: Assumes the active scalars will have a single component:
    vtkDataObject::SetPointDataActiveScalarInfo( ouInfo0, VTK_DOUBLE, 1 );
    }
  return 1;
}

int vtkDotSource::RequestUpdateExtent( vtkInformation* request, vtkInformationVector** inInfoVec, vtkInformationVector* ouInfoVec )
{
  return 1;
}

int vtkDotSource::RequestData( vtkInformation* request, vtkInformationVector** inInfoVec, vtkInformationVector* ouInfoVec )
{
  vtkPolyData* sim = this->SimulationData;
  if ( sim )
    {
    vtkInformation* ouInfo0 = ouInfoVec->GetInformationObject( 0 );
    if ( ! ouInfo0 )
      {
      return 0;
      }

    vtkDataObject* ouData0 = ouInfo0->Get( vtkDataObject::DATA_OBJECT() );
    if ( ouData0 )
      {
      ouData0->ShallowCopy( sim );
      ouData0->GetInformation()->Set( vtkDataObject::DATA_TIME_STEPS(), &this->SimulationTime, 1 );
      }
    }
  return 1;
}

void vtkDotSource::UpdateLocalParticles()
{
  double p = this->NumberOfTotalParticles;
  double n = this->NumberOfProcesses;
  double r = this->LocalId;
  this->LocalParticles[0] = static_cast<vtkIdType>( floor( r / n * p ) );
  this->LocalParticles[1] = static_cast<vtkIdType>( this->LocalId + 1 == this->NumberOfProcesses ? p : floor( (r + 1.) / n * p ) );
  this->NumberOfLocalParticles = this->LocalParticles[1] - this->LocalParticles[0];
  cout << "Process " << this->LocalId << " Particles " << this->LocalParticles[0] << " -- " << this->LocalParticles[1] << "\n";
}

