#ifndef __vtkDotSource_h
#define __vtkDotSource_h

#include "vtkPolyDataAlgorithm.h"

class vtkMultiProcessController;

class VTK_EXPORT vtkDotSource : public vtkPolyDataAlgorithm
{
public:
  static vtkDotSource* New();
  vtkTypeRevisionMacro(vtkDotSource,vtkPolyDataAlgorithm);
  virtual void PrintSelf( ostream& os, vtkIndent indent );

  // Description:
  // Return a pointer to the Dot data object.
  // This may return NULL, while New() may not.
  static vtkDotSource* GetSingleton() { return vtkDotSource::Singleton; }

  // Description:
  // This will destroy the source by releasing an additional reference on the singleton when it was created.
  // An error message will be reported if the reference count on the singleton is non-zero after this
  // supposedly-final reference is released.
  static void DestroySingleton();

  // Description:
  // Add an array to the simulation singleton dataset.
  virtual void AddPointArray( double* simStateData, int numComponents, const char* arrayName );

  // Description:
  // Prepare the singleton using data from the simulation.
  virtual void Initialize( int num_particles, double* pxyz, double* mass, double* pmom, double* pfrc, int local_id, int num_procs );

  // Description:
  // Has Initialize() been called at least once?
  virtual int IsInitialized() { return this->Initialized != 0; }

  // Description:
  // Update the time associated with the current output of the vtkDotSource filter.
  // This will cause a re-execution of the entire downstream pipeline when Update() is called.
  virtual void UpdateSimulationTime( double time );

  // Description:
  // Set/get a controller when the source is used in distributed-memory mode.
  virtual void SetController( vtkMultiProcessController* mpc );
  vtkGetObjectMacro(Controller,vtkMultiProcessController);

  // Description:
  // Get/set the rank of the local process.
  vtkSetMacro(LocalId,int);
  vtkGetMacro(LocalId,int);

  // Description:
  // Get/set the number of processes.
  vtkSetMacro(NumberOfProcesses,int);
  vtkGetMacro(NumberOfProcesses,int);

protected:
  vtkDotSource();
  virtual ~vtkDotSource();

  virtual int RequestInformation( vtkInformation* request, vtkInformationVector** inInfoVec, vtkInformationVector* ouInfoVec );
  virtual int RequestUpdateExtent( vtkInformation* request, vtkInformationVector** inInfoVec, vtkInformationVector* ouInfoVec );
  virtual int RequestData( vtkInformation* request, vtkInformationVector** inInfo, vtkInformationVector* ouInfo );

  void UpdateLocalParticles();

  vtkPolyData* SimulationData; // Container pointing to arrays owned by the simulation.
  double SimulationTime;
  int Initialized;
  vtkMultiProcessController* Controller;
  int LocalId;
  int NumberOfProcesses;
  vtkIdType LocalParticles[2];
  vtkIdType NumberOfLocalParticles;
  vtkIdType NumberOfTotalParticles;

private:
  vtkDotSource( const vtkDotSource& ); // Not implemented.
  void operator = ( const vtkDotSource& ); // Not implemented.

  static vtkDotSource* Singleton;
};

#endif // __vtkDotSource_h

