#include "vtkObjectFactory.h"
#include "vtkProcessModuleGUIHelper.h"
#include "vtkStdString.h"
#include "vtkstd/vector"
#include "vtkAlgorithm.h"
#include "vtkSMApplication.h"
#include "vtkSMProperty.h"
#include "vtkPVMain.h"
#include "vtkProcessModule.h"
#include "vtkPVPythonInterpretor.h"
#include "vtkPVPythonOptions.h"
#include "vtkInitializationHelper.h"
#include "vtkOpenGLRenderWindow.h"
#include "vtkMultiProcessController.h"

#include "vtkInsituPVBatch.h"

#include <vtksys/ios/sstream>

//---------------------------------------------------------------------------
class InsituPVBatchSingleton : public vtkProcessModuleGUIHelper
{
  static InsituPVBatchSingleton* gSingletonInstance;
  static InsituPVBatchSingleton* GetSingletonInstance ();

public:
  static InsituPVBatchSingleton *New ();
  vtkTypeRevisionMacro(InsituPVBatchSingleton, vtkProcessModuleGUIHelper);
  void PrintSelf(ostream& os, vtkIndent indent);

  void Initialize (vtkStdString script);
  void Finalize ();
  void SetDataSource (int index, vtkAlgorithm *source);
  vtkAlgorithm* GetDataSource (int index);
  void RunScript (vtkStdString script);
  void Update (int cycle, double time, double dt);
  
  // implementations of process module helper
  int RunGUIStart(int argc, char **argv, int numServerProcs, int myId); 
  int OpenConnectionDialog(int*) { return 1; }
  void SendPrepareProgress();
  void SendCleanupPendingProgress();
  void SetLocalProgress(const char*, int);
  void ExitApplication();

protected:
  InsituPVBatchSingleton ();
  ~InsituPVBatchSingleton ();

  vtkPVMain* PVMain;
  vtkPVPythonOptions* Options;

  vtkSMApplication* SMApplication;

  vtkPVPythonInterpretor* Interpretor;

  vtkstd::vector<vtkAlgorithm*> Sources;

private:
  InsituPVBatchSingleton (const InsituPVBatchSingleton&); // Not implemented
  void operator= (const InsituPVBatchSingleton&); // Not implemented
};

//---------------------------------------------------------------------------
vtkCxxRevisionMacro (InsituPVBatchSingleton, "$Revision: 1283 $");

InsituPVBatchSingleton* InsituPVBatchSingleton::gSingletonInstance = 0;

InsituPVBatchSingleton* InsituPVBatchSingleton::GetSingletonInstance ()
{
  if (InsituPVBatchSingleton::gSingletonInstance == 0)
    {
    InsituPVBatchSingleton::gSingletonInstance = 
          static_cast<InsituPVBatchSingleton*>(
              vtkObjectFactory::CreateInstance ("InsituPVBatchSingleton"));
    if (!InsituPVBatchSingleton::gSingletonInstance)
      {
      InsituPVBatchSingleton::gSingletonInstance = 
              new InsituPVBatchSingleton ();
      }
    }
  return InsituPVBatchSingleton::gSingletonInstance;
}

//---------------------------------------------------------------------------
InsituPVBatchSingleton* InsituPVBatchSingleton::New ()
{
  InsituPVBatchSingleton* ret = InsituPVBatchSingleton::GetSingletonInstance ();
  // ret->Register (0);
  return ret;
}

//---------------------------------------------------------------------------
InsituPVBatchSingleton::InsituPVBatchSingleton ()
{
  this->SMApplication = vtkSMApplication::New ();
  this->Interpretor = vtkPVPythonInterpretor::New();
}

//---------------------------------------------------------------------------
InsituPVBatchSingleton::~InsituPVBatchSingleton ()
{
  for (size_t i = 0; i < this->Sources.size (); i ++) 
    {
    this->Sources[i]->UnRegister (this);
    }
}

//---------------------------------------------------------------------------
void InsituPVBatchSingleton::PrintSelf (ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf (os, indent);
}

//---------------------------------------------------------------------------
static void ParaViewInitializeInterpretor (vtkProcessModule *pm)
{
  // Initialize built-in wrapper modules.
  vtkInitializationHelper::InitializeInterpretor (pm);
}

//---------------------------------------------------------------------------
void InsituPVBatchSingleton::Initialize (vtkStdString script)
{
  vtkOpenGLRenderWindow::SetGlobalMaximumNumberOfMultiSamples(0);

  int argc = 1;
  char **argv;
  argv = new char*[1];
  argv[0] = "pvinsitu";
  
  // Don't call MPI_Init in pvmain init because it was done by CTH
  //vtkPVMain::SetInitializeMPI (0); 
  vtkPVMain::Initialize (&argc, (char***)&argv);
  // Past pvmain init, remaining mpi inits assume MPI_Init already called
  //vtkPVMain::SetInitializeMPI (1); 

  this->PVMain = vtkPVMain::New ();
  this->Options = vtkPVPythonOptions::New ();
  this->Options->SetProcessType (vtkPVOptions::PVBATCH);
  this->Options->SetEnableSymmetricScripting (1);

  int ret = this->PVMain->Initialize (this->Options, this, 
                                ParaViewInitializeInterpretor, 
                                argc, argv);

  if (ret != 1)
    {
    vtkProcessModule::GetProcessModule()->SupportMultipleConnectionsOff ();
    this->Run (this->Options); // this is going to call RunGUIStart above
    }

  this->RunScript (script);

  delete [] argv;
}

//---------------------------------------------------------------------------
void InsituPVBatchSingleton::Finalize ()
{
  this->Interpretor->Delete ();
  this->SMApplication->Finalize ();
  this->SMApplication->Delete ();

  this->Options->Delete ();
  this->PVMain->Delete ();
  vtkPVMain::Finalize ();
}

//---------------------------------------------------------------------------
void InsituPVBatchSingleton::SetDataSource (int index, vtkAlgorithm *algorithm)
{
  if (this->Sources.size () <= index)
    {
    this->Sources.resize (index + 1);
    }
  this->Sources[index] = algorithm;
  this->Sources[index]->Register (this);
}

//---------------------------------------------------------------------------
vtkAlgorithm* InsituPVBatchSingleton::GetDataSource (int index)
{
  if (index >= this->Sources.size ())
    {
    vtkErrorMacro (<< "Data source index out of range");
    }
  return this->Sources[index];
}

//---------------------------------------------------------------------------
void InsituPVBatchSingleton::RunScript (vtkStdString script)
{
  this->Interpretor->RunSimpleString (script.c_str ());
}

//---------------------------------------------------------------------------
void InsituPVBatchSingleton::Update (int cycle, double time, double dt)
{
  vtkMultiProcessController *ctrl;
  ctrl = vtkMultiProcessController::GetGlobalController ();
  int id = ctrl->GetLocalProcessId ();
  vtksys_ios::ostringstream s;
  s << "update(" << id << "," << cycle << "," << time << "," << dt << ")" << endl;
  this->Interpretor->RunSimpleString (s.str ().c_str ());
}

//----------------------------------------------------------------------------
int InsituPVBatchSingleton::RunGUIStart(int argc, char **argv,
  int numServerProcs, int myId)
{
  (void)numServerProcs;

  this->SMApplication->Initialize();
  vtkSMProperty::SetCheckDomains(0);

  int res = 0; 

  vtkstd::vector<char*> vArg;
#define vtkPVStrDup(x) \
  strcpy(new char[ strlen(x) + 1], x)
  vArg.push_back(vtkPVStrDup(argv[0]));

  // Initializing the sub-interpreter because that is where RunSimpleString works.
  res = this->Interpretor->InitializeSubInterpretor(vArg.size(), &*vArg.begin());

  vtkstd::vector<char*>::iterator it;
  for ( it = vArg.begin(); it != vArg.end(); ++ it )
    {
    delete [] *it;
    }

  return res;
}

//----------------------------------------------------------------------------
void InsituPVBatchSingleton::SendPrepareProgress()
{
}

//----------------------------------------------------------------------------
void InsituPVBatchSingleton::SendCleanupPendingProgress()
{
}

//----------------------------------------------------------------------------
void InsituPVBatchSingleton::SetLocalProgress(const char*, int)
{
}

//----------------------------------------------------------------------------
void InsituPVBatchSingleton::ExitApplication()
{ 
}

//----------------------------------------------------------------------------

vtkCxxRevisionMacro (vtkInsituPVBatch, "$Revision: 1283 $");
vtkStandardNewMacro (vtkInsituPVBatch);

vtkInsituPVBatch::vtkInsituPVBatch ()
{
  this->implementation = InsituPVBatchSingleton::New ();
}

vtkInsituPVBatch::~vtkInsituPVBatch ()
{
  this->implementation->Delete ();
}

void vtkInsituPVBatch::PrintSelf (ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf (os, indent);
}


void vtkInsituPVBatch::Initialize (vtkStdString script)
{
  this->implementation->Initialize (script);
}

void vtkInsituPVBatch::Finalize ()
{
  this->implementation->Finalize ();
}

void vtkInsituPVBatch::SetDataSource (int index, vtkAlgorithm *source)
{
  this->implementation->SetDataSource (index, source);
}

vtkAlgorithm *vtkInsituPVBatch::GetDataSource (int index)
{
  return this->implementation->GetDataSource (index);
}

void vtkInsituPVBatch::RunScript (vtkStdString script)
{
  return this->implementation->RunScript (script);
}

void vtkInsituPVBatch::Update (int cycle, double time, double dt)
{
  this->implementation->Update (cycle, time, dt);
}

